#include <strings.h>
#include "FECReceiver.h"

TFECReceiver::TFECReceiver(TGlobals *gIn) :
  TReceiver(gIn)
{
  bzero(que, sizeof(double) * QUE_SIZE);
  setupInitBuffer();
}

//called when calculate the mu for the playout Delay. 
void TFECReceiver::getRecentNetworkLossRate()
{
  int count = 0;

  for (int i=0; i< HISTORY_BUFF_SIZE; i++)
  {
    //0: networkLoss. 4: netowrkLoss but recovered.
    if(history_buffer[i] == 0 || history_buffer[i] == 4)
      count++;
  }

  recentLossRate = ((double)count)/HISTORY_BUFF_SIZE;

}

//read 6 packets in que, we need to look forward
void TFECReceiver::setupInitBuffer()
{
  TReceiver::readOneTrace();
  que[(prevSeqId-1)%QUE_SIZE] = networkDelay;

  if (!g->isMaxDelay)
  {
    playoutDelay = networkDelay + INIT_RELATIVE_DELAY;
  }

  for(int i=1; i<6; i++)
  {
    TReceiver::readOneTrace();
    que[(prevSeqId-1)%QUE_SIZE] = networkDelay;
  }
}

bool TFECReceiver::isNetworkOrDelayLoss()
{
  TReceiver::readOneTrace();
  que[(prevSeqId-1)%QUE_SIZE] = networkDelay;

  //remember we are always 6 packets later
  int curSeqId = prevSeqId - 6;
  int curPos = (curSeqId - 1)%QUE_SIZE;

  if (endFlag && buf[0][1] != RATE_8000 &&
        buf[1][1] != RATE_8000 && buf[2][1] != RATE_8000)
  { //null packet, can't contribute in recovering.
    que[curPos] = -1;
    return false;
  }

  int count=0;

  if (que[curPos]==0 || que[curPos]>playoutDelay)
  {
    int pos = (curSeqId - spurtStart) % 3;
    int startTmpPos = (curSeqId - 1)%QUE_SIZE - pos;
    int startPos = (startTmpPos >= 0)? startTmpPos : startTmpPos + QUE_SIZE;

    //see if we can recover it.
    for (int i=0; i<5; i++)
    {
      int calcPos = (startPos+i)%QUE_SIZE;

      if( que[calcPos]==0 || que[calcPos]==-1 || que[calcPos]>playoutDelay)
      {
        sortArea[i] = 0;
      } else {
        count ++;
        sortArea[i] = que[calcPos] + g->numFramePerPacket * 10 * i;
      }
    }

    if (count > 2)
    {
      que[curPos] = getVirtualNetworkDelay() 
                      - g->numFramePerPacket * 10 * (curPos-startPos);
      totalRecovered ++;
    }
  }

  if (que[curPos] != 0 && que[curPos] != -1 && que[curPos]<playoutDelay)
  {
    networkDelay = que[curPos];
    updateMeanDelay();
    if (count != 0)
    {
      history_buffer[(prevSeqId-1)%HISTORY_BUFF_SIZE] += 4;
    }
    return false;
  } else {
    if (que[curPos] == 0)
    {
      fprintf(g->fLog, "%d: Network Loss.\n", curSeqId);
    } else if (que[curPos] > 0) {
      fprintf(g->fLog, "%d: Late Loss.\n", curSeqId);
      history_buffer[(prevSeqId-1)%HISTORY_BUFF_SIZE] += 2;
      totalDelayLoss ++;
    }

    fflush(g->fLog);
    return true;
  }
}

double TFECReceiver::getVirtualNetworkDelay()
{ //we need to have non-zero and the third smallest as our VirtualDelay;

  int pt = -1;

  for (int i=0; i<3; i++)
  {
    for (int j=0; j<5; j++)
    {
      if(sortArea[j] != 0)
      {
        if (pt == -1 || sortArea[j] < sortArea[pt])
          pt = j;
      }
    }
    if (i!=2)
    {
      sortArea[pt] = 0;
      pt = -1;
    } 
  }

  return sortArea[pt];
}

void TFECReceiver::printDelayInfo()
{
  fprintf(g->fLog, "%d:%d:%dms: PlayoutDelay: %.4lf, adjustDelay: %.4lf\n", prevSeqId-6, (prevSeqId-6-startSeqId), (prevSeqId-7-startSeqId)*10*g->numFramePerPacket, playoutDelay,  adjustDelay);
  fflush(g->fLog);

}
