#include "Receiver.h"
#include <strings.h>

TReceiver::TReceiver(TGlobals *gIn)
{
  Init(gIn);
}

TReceiver::~TReceiver()
{
  delete coder;
  delete decoder;
}

void TReceiver::Init(TGlobals *gIn)
{
  //allocate from here. Otherwise, can't pass through (1) in the header file
  coder = new g729_encoder_basic(1); 
  decoder = new g729_decoder_basic;

  g = gIn;

  totalNumPacket = 0;
  totalNetworkLoss = 0;
  totalDelayLoss = 0;
  totalRecovered = 0;
  totalPlayoutDelay = 0.0;
  recentLossRate = 0.0;

  if (gIn->isSpike)
  {
    alpha = 0.875;
  } else {
    alpha = 0.998002;
  }
  mu = 4.0;
  prevSeqId = 0;
  seqId = 0;
  startSeqId = 1;
  var = 0;
  variant = 10;

  muMin = 0;
  muMax = 8;
  seta = 0.05;
  detaIncrease = 0.4;
  detaDecrease = 0.2;

  initialized = true;
  endFlag = false;
  networkDelay = 0;
  prevNetworkDelay = 0;
  prevPrevNetworkDelay = 0;
  meanDelay = 0;
  maxDelay = 0;
  playoutDelay = 0;
  prevPlayoutDelay = 0;
  spikeMode = NORMAL;
  delay = 0;
  frameCount = 1;
  spurtStart = 1;

  adjustDelay = 0; //delay between current and prvious delay

  g->fIn = fopen(g->inputFile, "rb");
  g->fOut = fopen(g->outputFile, "w+b");
  g->fLoss = fopen(g->lossFile, "w+b");
  g->fTrace = fopen(g->traceFile, "r");
  g->fLog = fopen(g->logFile, "w");
  g->fVad = fopen(g->vadFile, "w");

  if (g->fLoss == NULL || g->fIn == NULL || g->fTrace == NULL || 
        g->fOut == NULL || g->fLog == NULL || g->fVad == NULL)
  {
    g->handleExit("Unable to open files.\n");
  }

  if (g->isMaxDelay)
  {
    fTmp = fopen(".tmp", "wb");
    fTmpLoss = fopen(".tmpLoss", "wb");
    if (fTmp == NULL || fTmpLoss == NULL)
    {
      g->handleExit("Unable to open tmp files.\n");
    }

    maxInMaxDelay = 500;
    forceCalcFlag = false;
    numNullFrame = -1;
  }

  //now set the starting read line for the trace file
  char line[100];

  for (int i=0; i<g->lineSkip; i++)
  {
    fgets(line, 100, g->fTrace);
  }
 
  if (g->lineSkip > 0)
  {
    if (sscanf(line, "%d %lf", &startSeqId, &delay) < 2)
    {
      g->handleExit("network delay wrong input format");
    }
    prevSeqId = startSeqId;
    seqId = startSeqId;
  }

  //mid-spurt
  oneDelayOnFlag = false;
  midSpurtSeta = 2.0;
}

//main function called by outside
void TReceiver::generate()
{
  if (g->isMaxDelay)
  {
    while(1)
    {
      processMaxDelayPacket();
    }
  } else if (g->midSpurtFlag) {
    while (1)
    {
      processMidSpurtPacket();
    }
  } else {
    while (1)
    {
      processPacket();
    }
  }
}

void TReceiver::processMaxDelayPacket()
{
  int len;
  INT16 silence[SAMPLE_PER_FRAME+10];

  bzero(silence, sizeof(INT16) * (SAMPLE_PER_FRAME+10));
  silence[0] = 27425;
  silence[1] = RATE_0;

  //read packet from input file. 
  len = readOnePacket();

  //first calculate its loss and delay information
  if (isNetworkOrDelayLoss())
  {
    //if it is networkLoss, write background noise
    for (int i=0; i<len; i++)
    {
      bzero(&(buf[i][2]), sizeof(INT16) * buf[i][1]);
      fwrite(buf[i], sizeof(INT16), SAMPLE_PER_FRAME+10, fTmp);

      //write loss waves
      for (int k=0; k<SAMPLE_PER_FRAME; k++)
      {
        buf[i][k] = 10000;
      }

      fwrite(buf[i], sizeof(INT16), SAMPLE_PER_FRAME, fTmpLoss);
    }

    return;
  }

  //process each frame within a packet
  for (int i=0; i< len; i++)
  {
    fwrite(buf[i], sizeof(INT16), SAMPLE_PER_FRAME+10, fTmp);
    fwrite(silence+2, sizeof(INT16), SAMPLE_PER_FRAME, fTmpLoss);

    if (isNormalFrame(buf[i]) && endFlag )
    {
      spurtStart = prevSeqId;
    }

    //only the first ending packet will be marked as spurt end package
    if (isSpurtEnd(buf[i]))
    {
      writePackets(silence);
      //now, reset null frame
      numNullFrame = -1;
      fTmp = fopen(".tmp", "wb");
      fTmpLoss = fopen(".tmpLoss", "wb");
      //reset playoutDelay
      delayInMaxDelay = 0;
    }

    if (!isNormalFrame(buf[i]) && endFlag )
    {
      numNullFrame ++;
    }
  }

  if (forceCalcFlag)
  {
    writePackets(silence);
  }
}

void TReceiver::writePackets(INT16* silence)
{

        //end of the talk spurt, update playout delay information.
        prevPlayoutDelay = (playoutDelay == 0)? delayInMaxDelay : playoutDelay;
        playoutDelay = delayInMaxDelay;

        //should we suppress or add more background noise?
        adjustDelay = getRelativePlayoutDelay();

        printDelayInfo();

        fclose(fTmp);
        fclose(fTmpLoss);
        fTmp = fopen(".tmp", "rb");
        fTmpLoss = fopen(".tmpLoss", "rb");

        //Now, adjust playoutDelay
        if (adjustDelay > 0)
        { //we need extra silence, add them
          while (adjustDelay > 0)
          {
            double extraDelay = (adjustDelay > 10) ? 10 : adjustDelay;

            int numSilence = (int) (SAMPLE_PER_FRAME * extraDelay/10);

            decoder->decode_1_frame(silence);
            fwrite16(decoder->pst_out, numSilence, g->fOut);
            fwrite(silence+2, sizeof(INT16), numSilence, g->fLoss);

            adjustDelay = adjustDelay - 10;
          }

          adjustDelay = 0;
        }
 
        while (adjustDelay < 0 && numNullFrame > 0)
        { // we need to compress the silence
          if (adjustDelay < -10)
          {
            adjustDelay = adjustDelay + 10;
          } else {
            //only compress some
            int numSilence = SAMPLE_PER_FRAME
                             - (int) (SAMPLE_PER_FRAME * fabs(adjustDelay)/10);

            decoder->decode_1_frame(silence);
            fwrite16(decoder->pst_out, numSilence, g->fOut);

            fwrite(silence+2, sizeof(INT16), numSilence, g->fLoss);

            adjustDelay = 0;
          }

          numNullFrame --;
          fread(buf[0], sizeof(INT16), SAMPLE_PER_FRAME, fTmpLoss);
          fread(buf[1], sizeof(INT16), SAMPLE_PER_FRAME+10, fTmp);
        }

        if (adjustDelay < 0 )
        { //zero out the adjust Delay so that it won't carry balance over
          playoutDelay = playoutDelay - adjustDelay;
          fprintf(g->fLog, "\t\tGaps can't shrink: %.4lf. Current playoutDelay:%.4lf\n", adjustDelay, playoutDelay);
          adjustDelay = 0;
          spurtStart = prevSeqId;
        }

        //write out the rest of the frames
        while (fread(buf[1], sizeof(INT16), SAMPLE_PER_FRAME+10, fTmp) > 0)
        {
          decoder->decode_1_frame(buf[1]);
          fwrite16(decoder->pst_out, SAMPLE_PER_FRAME, g->fOut);

          fread(buf[0], sizeof(INT16), SAMPLE_PER_FRAME, fTmpLoss);
          fwrite(buf[0], sizeof(INT16), SAMPLE_PER_FRAME, g->fLoss);
        }

        fclose(fTmp);
        fclose(fTmpLoss);
}

bool TReceiver::isNetworkOrDelayLoss()
{
  if (isNetworkLoss())
  {
    fprintf(g->fLog, "%d: Network Loss.\n", prevSeqId);
    fflush(g->fLog);

    return true;
  }

  //null packets are not transmitted in real situation. Therefore there is
  //no packet loss
  if (endFlag &&
      buf[0][1] != RATE_8000 && buf[1][1] != RATE_8000 && buf[2][1] != RATE_8000
)
  {
    return false;
  }

  if (g->isMaxDelay)
  {
    if (networkDelay < maxInMaxDelay)
    {
      delayInMaxDelay = (delayInMaxDelay > networkDelay) ?
                                              delayInMaxDelay : networkDelay;
      return false;
    } else {
      delayInMaxDelay = (delayInMaxDelay > maxInMaxDelay) ?
                                              maxInMaxDelay : delayInMaxDelay;
      totalDelayLoss ++;
      fprintf(g->fLog, "%d: Late Loss.\n", prevSeqId);
      fflush(g->fLog);

      return true;
    }
  }

  //notice that lateLoss only applies to packets transmitted
  if (isLateLoss())
  {
    fprintf(g->fLog, "%d: Late Loss.\n", prevSeqId);
    fflush(g->fLog);

    return true;
  }

  return false;
}

void TReceiver::processMidSpurtPacket()
{
  int len;
  
  //read one packet from input file.
  len = readOnePacket();

  if (isNetworkOrDelayLoss())
  {
    //notice that for null frame, it won't transmit. So the only situation that 
    //we need to pay attention to is that first late packet is normal, and 
    //the second late packet contains talk spurt. Then we will have no problem
    //if we need to adjust twice, one for mid spurt adjust, and one for regular
    //spurt end adjust. 
    if (networkDelay != 0)
    { //delay loss
      if (oneDelayOnFlag == false)
      { //first delay, ignore
        oneDelayOnFlag = true;
        bcopy(buf[0], midSpurtBuf[0], sizeof(INT16) * 3 * (SAMPLE_PER_FRAME+10));
        return;
      } else {
        //second delay
        oneDelayOnFlag = false;

        totalDelayLoss = totalDelayLoss -2;
        //write the previous packet
        bcopy(buf[0], midSpurtBuf[3], sizeof(INT16) * 3 * (SAMPLE_PER_FRAME+10));
        buf[0][0] = midSpurtBuf[0][0];
        buf[0][1] = RATE_8000;
        bzero(&(buf[0][2]), sizeof(INT16) * SAMPLE_PER_FRAME);
        bzero(buf[1], sizeof(INT16) * (SAMPLE_PER_FRAME+10));

        double tmpDelay = (networkDelay > prevPrevNetworkDelay)? networkDelay : prevPrevNetworkDelay;
        double adjustMidDelay = tmpDelay - playoutDelay + midSpurtSeta;

        //now, playoutDelay is larger
        playoutDelay = playoutDelay + adjustMidDelay;

        fprintf(g->fLog, "%d:%d:%dms: PlayoutDelay: %.4lf, midSpurtAdjustDelay: %.4lf\n", prevSeqId, (prevSeqId-startSeqId), (prevSeqId-1-startSeqId)*10*g->numFramePerPacket, playoutDelay,  adjustMidDelay);
        fflush(g->fLog);

        //add bad packets
        while (adjustMidDelay > 0)
        {
          double extraDelay = (adjustMidDelay > 10) ? 10 : adjustMidDelay;

          int numSilence = (int) (SAMPLE_PER_FRAME * extraDelay/10);

          decoder->decode_1_frame(buf[0]);
          fwrite16(decoder->pst_out, numSilence, g->fOut);
          fwrite(buf[1], sizeof(INT16), numSilence, g->fLoss);

          adjustMidDelay = adjustMidDelay - 10;
        }

        bcopy(midSpurtBuf[0], buf[0], sizeof(INT16) * 3 * (SAMPLE_PER_FRAME+10));
        handlePacket(len);
        bcopy(midSpurtBuf[3], buf[0], sizeof(INT16) * 3 * (SAMPLE_PER_FRAME+10));
        handlePacket(len);
      }
    } else {
      if (oneDelayOnFlag)
      {
        oneDelayOnFlag = false;
        writeLossPacket(g->numFramePerPacket);
      }

      //if it is networkLoss, write background noise
      writeLossPacket(len);
    }

  } else {
    if (oneDelayOnFlag)
    {
      oneDelayOnFlag = false;

      writeLossPacket(g->numFramePerPacket);

    }

    handlePacket(len);
  }
}

void TReceiver::writeLossPacket(int length)
{
  INT16 buffer[SAMPLE_PER_FRAME+10];
  INT16 lossBuffer[SAMPLE_PER_FRAME];

  //write it as a loss
  bzero(buffer, sizeof(INT16) * (SAMPLE_PER_FRAME+10));
  buffer[0] = 27425;
  buffer[1] = RATE_8000;
  decoder->decode_1_frame(buffer);

  //write loss waves
  for (int k=0; k<SAMPLE_PER_FRAME; k++)
  {
    lossBuffer[k] = 10000;
  }

  for (int i=0; i< length; i++)
  {
    fwrite16(decoder->pst_out, SAMPLE_PER_FRAME, g->fOut);
    fwrite(lossBuffer, sizeof(INT16), SAMPLE_PER_FRAME, g->fLoss);
  }
}

//main function that handles each packet
void TReceiver::processPacket()
{
  int len;

  //read packet from input file. 
  len = readOnePacket();

  //first calculate its loss and delay information
  if (isNetworkOrDelayLoss())
  {
    //if it is networkLoss, write background noise
    for (int i=0; i<len; i++)
    {
      bzero(&(buf[i][2]), sizeof(INT16) * buf[i][1]);
      decoder->decode_1_frame(buf[i]);
      fwrite16(decoder->pst_out, SAMPLE_PER_FRAME, g->fOut);

      //write loss waves
      for (int k=0; k<SAMPLE_PER_FRAME; k++)
      {
        buf[i][k] = 10000;
      }

      fwrite(buf[i], sizeof(INT16), SAMPLE_PER_FRAME, g->fLoss);
    }

    return;
  }

  handlePacket(len);
}

void TReceiver::handlePacket(int len)
{
  INT16 silence[SAMPLE_PER_FRAME+10];

  bzero(silence, sizeof(INT16) * (SAMPLE_PER_FRAME+10));
  silence[0] = 27425;
  silence[1] = RATE_0;

  //if there are two talk spurt within one packet, just treat them as one 
  //talk spurt
  int spurtCount = (adjustDelay != 0)? 1 : 0;

  //process each frame within a packet
  for (int i = 0; i < len; i++)
  {
    //only the first ending packet will be marked as spurt end package
    if (isSpurtEnd(buf[i]))
    {
      if (++spurtCount > 1)
      {
        endFlag = false;
      } else {
        //end of the talk spurt, update playout delay information.
        calculatePlayoutDelay();
        //should we suppress or add more background noise?
        adjustDelay = getRelativePlayoutDelay();

        printDelayInfo();
      }
    }

    //beginning of the talk spurt
    if (adjustDelay > 0 && isNormalFrame(buf[i]))
    { //we need extra silence, add them
      while (adjustDelay > 0)
      {
        double extraDelay = (adjustDelay > 10) ? 10 : adjustDelay;

        int numSilence = (int) (SAMPLE_PER_FRAME * extraDelay/10);

        decoder->decode_1_frame(silence);
        fwrite16(decoder->pst_out, numSilence, g->fOut);
        fwrite(silence+2, sizeof(INT16), numSilence, g->fLoss);

        adjustDelay = adjustDelay - 10;
      }

      adjustDelay = 0;
      spurtStart = prevSeqId;
    }
  
    if (adjustDelay < 0 && isNormalFrame(buf[i]))
    { //zero out the adjust Delay so that it won't carry balance over
      playoutDelay = playoutDelay - adjustDelay;
      fprintf(g->fLog, "\t\tGaps can't shrink: %.4lf. Current playoutDelay:%.4lf\n", adjustDelay, playoutDelay);
      adjustDelay = 0;
      spurtStart = prevSeqId;
    }

    if (adjustDelay < 0 && isNullFrame(buf[i]))
    { // we need to compress the silence

      if (adjustDelay < -10)
      {
        adjustDelay = adjustDelay + 10;
      } else {
        //only compress some
        int numSilence = SAMPLE_PER_FRAME 
                             - (int) (SAMPLE_PER_FRAME * fabs(adjustDelay)/10);

        decoder->decode_1_frame(buf[i]);
        fwrite16(decoder->pst_out, numSilence, g->fOut);

        fwrite(silence+2, sizeof(INT16), numSilence, g->fLoss);

        adjustDelay = 0;
      } 
    } else {
      //regular cases, just write the packets
      decoder->decode_1_frame(buf[i]);
      fwrite16(decoder->pst_out, SAMPLE_PER_FRAME, g->fOut);

      fwrite(silence+2, sizeof(INT16), SAMPLE_PER_FRAME, g->fLoss);
    }
  }
}

void TReceiver::printDelayInfo()
{
  fprintf(g->fLog, "%d:%d:%dms: PlayoutDelay: %.4lf, adjustDelay: %.4lf\n", prevSeqId, (prevSeqId-startSeqId), (prevSeqId-startSeqId-1)*10*g->numFramePerPacket, playoutDelay,  adjustDelay);
  fflush(g->fLog);

}

//read from sound file one packet
int TReceiver::readOnePacket()
{
  //read the sound
  //codec
  INT16 frame[SAMPLE_PER_FRAME];
  int count;
  int i;

  for (count=0; count < g->numFramePerPacket; count++)
  {
    if ( (i=fread(frame, sizeof(INT16), SAMPLE_PER_FRAME, g->fIn)) > 0)
    {
      coder->encode_1_frame(frame);

      fprintf(g->fVad,"%d: %d\n", frameCount, coder->serial[1]);
      fflush(g->fVad);
      frameCount++;

      memcpy(buf[count], coder->serial, 2 * (coder->serial[1] +2));
    } else if (i < 0) {
      g->handleExit("Read Error for input file.\n");
    } else {
      break;
    }
  }

  if( count == 0)
  {
    if (g->isMaxDelay && forceCalcFlag == false)
    {
      forceCalcFlag = true;
      return 0;
    }

    fprintf(g->fLog, "TotalPacket: %d\nTotalNetworkLoss: %d\nTotalDelayLoss: %d\nTotalRecovered: %d\nAvg playout Delay: %.4lf\n", totalNumPacket, totalNetworkLoss, totalDelayLoss, totalRecovered, totalPlayoutDelay/totalNumPacket);
    fflush(g->fLog);

    g->handleExit("Finished.\n");
  }

  return count;
}

//read from the trace file packet information. Calculate the mean delay. Mark
//the loss situation
bool TReceiver::isNetworkLoss()
{
  readOneTrace();

  if(prevSeqId == seqId)
  {
    updateMeanDelay();
  }

  if (networkDelay == 0)
    return true;
  else
    return false;
}

void TReceiver::readOneTrace()
{
  char line[100];

  if (prevSeqId == seqId)
  {
    fgets(line, 100, g->fTrace);
    if (sscanf(line, "%d %lf", &seqId, &delay) < 2)
    {
      fprintf(g->fLog, "network delay wrong input format");
    }
  }

  if (++prevSeqId == seqId)
  {
    networkDelay = 1000 * delay;  //convert it to ms
    history_buffer[(prevSeqId-1)%HISTORY_BUFF_SIZE] = 1;

  } else {
    networkDelay = 0;
    totalNetworkLoss ++;
    history_buffer[(prevSeqId-1)%HISTORY_BUFF_SIZE] = 0;
  }
}

bool TReceiver::isLateLoss()
{
  if (playoutDelay < networkDelay)
  {
    totalDelayLoss ++;
    history_buffer[(prevSeqId-1)%HISTORY_BUFF_SIZE] += 2;
    return true;
  } else {
    return false;
  }
}

void TReceiver::updateMeanDelay()
{
  if (! (endFlag && buf[0][1] != RATE_8000 && 
        buf[1][1] != RATE_8000 && buf[2][1] != RATE_8000))
  {
    calculateMeanDelay();
  } 
}

void TReceiver::calculateMeanDelay()
{
  totalNumPacket ++;
  //now update the meanDelay

  if (meanDelay == 0)
  {
    meanDelay = networkDelay;
    if (!g->isMaxDelay)
    {
      playoutDelay = networkDelay + INIT_RELATIVE_DELAY; 
    }

    maxDelay = networkDelay;
    spikeMode = NORMAL;
    prevNetworkDelay = networkDelay;
    prevPrevNetworkDelay = networkDelay;
  }

  if (g->isPrevOpt) {
    maxDelay = (maxDelay > networkDelay) ? maxDelay : networkDelay;
    meanDelay = 0.25 * meanDelay + 0.75 * maxDelay;
  } else if (g->isSpike) {
    if (spikeMode == NORMAL)
    {
      if (fabs(networkDelay - prevNetworkDelay) > fabs(variant) * 2 + TO_SPIKE_THRESHOLD)
      { 
        var = 0;
        spikeMode = IMPULSE;
      }

    } else {
      var = var/2 + 
            fabs(2 * networkDelay - prevNetworkDelay - prevPrevNetworkDelay)/8;
      if (var <= TO_NORMAL_THRESHOLD)
      {
        spikeMode = NORMAL;
      }
    }

    if (spikeMode == NORMAL) {
      meanDelay = 0.125 * networkDelay + 0.875 * meanDelay;
    } else {
      meanDelay = meanDelay + networkDelay - prevNetworkDelay;
    }

  } else { //exp or nothing is specified
    meanDelay = alpha * meanDelay + (1 - alpha) * networkDelay;
  }

  //I am using the same variant for every algorithm
  variant = alpha * variant + (1 - alpha) * fabs(meanDelay - networkDelay);

  totalPlayoutDelay = playoutDelay + totalPlayoutDelay;

  prevPrevNetworkDelay = prevNetworkDelay;
  prevNetworkDelay = networkDelay;
}

//this packet returns the first indication of SID or NULL frame
bool TReceiver::isSpurtEnd(INT16 *frame)
{
  int frameBytes = (int)frame[1];

  if (frameBytes == RATE_8000 && endFlag)
  { //first packet
    endFlag = false;
  } else if ( (frameBytes == RATE_SID_OCTET || frameBytes == RATE_SID ||
               frameBytes == RATE_0) &&
                     endFlag == false) {
    endFlag = true;
    return true;
  }

  return false;
}

bool TReceiver::isNullFrame(INT16 * frame)
{
  int frameBytes = (int)frame[1];

  return (frameBytes == RATE_0) ? true : false;
}

bool TReceiver::isNormalFrame(INT16 * frame)
{
  int frameBytes = (int)frame[1];

  return (frameBytes == RATE_8000) ? true : false;
}

void TReceiver::getRecentNetworkLossRate()
{
  int count = 0;
  int denom = HISTORY_BUFF_SIZE;

  if (prevSeqId - startSeqId < HISTORY_BUFF_SIZE)
  {
    denom = prevSeqId - startSeqId;
  }

  for (int i=0; i< denom; i++)
  {
    if(history_buffer[i] == 0)
      count++;
  }

  recentLossRate = ((double)(count))/denom;
}

double TReceiver::getRecentNetworkAppLossRate()
{
  int count = 0;
  int denom = HISTORY_BUFF_SIZE;

  if (prevSeqId - startSeqId < HISTORY_BUFF_SIZE)
  {
    denom = prevSeqId - startSeqId;
  }

  for (int i=0; i<denom; i++)
  {
    if(history_buffer[i] == 0 || history_buffer[i] == 3)
      count++;
  }

  return ((double)count)/denom;
}

//this function is called only at a spurt
void TReceiver::calculatePlayoutDelay()
{
  prevPlayoutDelay = playoutDelay;

  getRecentNetworkLossRate();

  double PC = 0.0;
  double PL = 0.0;

  if (g->fecFlag)
  {
    //suppose our target loss rate = 0
    PC =  1- 4 * pow((1-recentLossRate), 3) * recentLossRate 
                                                  - pow((1-recentLossRate), 4);
  } else {
    PC = recentLossRate;
  }

  PL = getRecentNetworkAppLossRate();
    
  if ((PC < PL - seta)  && (mu <= muMax - detaIncrease))
  {
    mu = mu + detaIncrease;
  } else if ( (PC > PL + seta) && (mu >= muMin + detaDecrease)) {
    mu = mu - detaDecrease;
  } 
  fprintf(g->fLog, "mu=%.4lf\t", mu);

  playoutDelay = meanDelay + mu * variant; 
}

double TReceiver::getRelativePlayoutDelay()
{
/*
  if (networkDelay > playoutDelay) //this packet itself already late
  {
    return (playoutDelay - prevPlayoutDelay - (networkDelay - playoutDelay));
  } else {
    return (playoutDelay - prevPlayoutDelay);
  }
*/
  return (playoutDelay - prevPlayoutDelay);

}
