#ifndef _RECEIVER_H
#define _RECEIVER_H

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "g729_encoder.h"
#include "g729_decoder.h"
#include "Globals.h"

class TReceiver
{
  public:
    void generate();
    TReceiver(TGlobals *gIn);
    ~TReceiver();

    enum {
      TO_SPIKE_THRESHOLD = 20,
      TO_NORMAL_THRESHOLD = 5,
      HISTORY_BUFF_SIZE = 200
    };

  protected:
    virtual bool isNetworkOrDelayLoss();
    void processPacket();
    void Init(TGlobals *gIn);
    virtual int readOnePacket();
    void readOneTrace();
    bool isNetworkLoss();
    bool isLateLoss();
    bool isSpurtEnd(INT16 *frame);
    bool isNullFrame(INT16 * frame);
    bool isNormalFrame(INT16 * frame);
    virtual void calculatePlayoutDelay();
    double getRelativePlayoutDelay();
    virtual void updateMeanDelay();
    void calculateMeanDelay();
    virtual void getRecentNetworkLossRate();
    virtual void printDelayInfo();
    double getRecentNetworkAppLossRate();
    void processMaxDelayPacket();
    void processMidSpurtPacket();
    void writePackets(INT16* silence);
    void handlePacket(int len);
    void writeLossPacket(int length);

    TGlobals *g;

    g729_encoder_basic *coder; //1: use VAD/DTX (silence detection)
    g729_decoder_basic *decoder;

    INT16 buf[MAX_NUM_FRAME_PER_PACKET][SAMPLE_PER_FRAME+10]; //2 extra usually
    int history_buffer[HISTORY_BUFF_SIZE];

    //adjust delay is only for every spurt. And it does not accumulate
    double adjustDelay;
    int totalNumPacket;
    int totalDelayLoss;
    int totalNetworkLoss;
    int totalRecovered;
    int prevSeqId;
    int startSeqId;
    int seqId;
    int spurtStart;
    int frameCount;
    double alpha;
    double mu;
    double totalPlayoutDelay;
    double recentLossRate;

    bool initialized;
    bool endFlag;
    double networkDelay; //in ms
    double delay;
    double prevNetworkDelay;
    double prevPrevNetworkDelay;
    double meanDelay;
    double playoutDelay;
    double prevPlayoutDelay;
    double maxDelay;
    double var;  //variable evaluating ending of the spike mode
    double variant;

    double muMin;
    double muMax;
    double seta;
    double detaIncrease;
    double detaDecrease;

    enum TSpikeMode {IMPULSE, NORMAL};
    TSpikeMode spikeMode;

    //MaxDelay
    double maxInMaxDelay;
    double delayInMaxDelay;
    FILE* fTmp;
    FILE* fTmpLoss;
    bool forceCalcFlag;
    int numNullFrame;

    //for midSpurt
    INT16 midSpurtBuf[6][SAMPLE_PER_FRAME+10]; 
    bool oneDelayOnFlag;
    double midSpurtSeta;
};
#endif
